@@ -10,6 +10,7 @@
 #include <sys/time.h>
 #include <time.h>
 #include "jdupes.h"
+#include "loaddir.h"
 #include "libjodycode.h"
 #include "likely_unlikely.h"
 #include "hashdb.h"
@@ -27,6 +28,10 @@
 #endif
 #define HT_MASK (HT_SIZE - 1)
 
+#if defined(__GNU__) && !defined(PATH_MAX)
+ #define PATH_MAX 1024
+#endif
+
 static hashdb_t *hashdb[HT_SIZE];
 static int hashdb_init = 0;
 static int hashdb_algo = 0;
@@ -72,7 +77,7 @@ int save_hash_database(const char * const restrict dbname, const int destroy)
   char *dbtemp;
 
   if (dbname == NULL) goto error_hashdb_null;
-  LOUD(fprintf(stderr, "save_hash_database('%s')\n", dbname);)
+  LOUD(fprintf(stderr, "save_hash_database('%s') destroy = %d, dirty = %d\n", dbname, destroy, hashdb_dirty);)
   /* Don't save the hash database if it wasn't changed */
   if (hashdb_dirty == 0 && destroy == 0) return 0;
   if (hashdb_dirty == 1) {
@@ -81,22 +86,18 @@ int save_hash_database(const char * const restrict dbname, const int destroy)
     strcpy(dbtemp, dbname);
     strcat(dbtemp, ".tmp");
     /* Try to remove any existing temporary database, ignoring errors */
-    remove(dbtemp);
-    errno = 0;
-    db = fopen(dbtemp, "wb");
+    jc_remove(dbtemp);
+    db = jc_fopen(dbtemp, JC_FILE_MODE_RW_SEQ);
     if (db == NULL) goto error_hashdb_open;
-  }
-
-  if (hashdb_dirty == 1) {
     if (write_hashdb_entry(db, NULL, &cnt, destroy) != 0) goto error_hashdb_write;
-    fclose(db);
+    jc_fclose(db);
     if (new_hashdb == 0) {
-      errno = 0;
-      if (remove(dbname) != 0) {
-        if (errno != ENOENT) goto error_hashdb_remove;
+      jc_errno = 0;
+      if (jc_remove(dbname) != 0) {
+        if (jc_errno != ENOENT) goto error_hashdb_remove;
       }
     }
-    if (rename(dbtemp, dbname) != 0) goto error_hashdb_rename;
+    if (jc_rename(dbtemp, dbname) != 0) goto error_hashdb_rename;
     LOUD(if (hashdb_dirty == 1) fprintf(stderr, "Wrote %" PRIu64 " items to hash databse '%s'\n", cnt, dbname);)
     hashdb_dirty = 0;
   }
@@ -111,14 +112,14 @@ error_hashdb_open:
   return -2;
 error_hashdb_write:
   fprintf(stderr, "error: write failed to temp hashdb '%s': %s\n", dbtemp, strerror(errno));
-  fclose(db);
+  jc_fclose(db);
   return -3;
 error_hashdb_alloc:
   fprintf(stderr, "error: cannot allocate memory for temporary hashdb name\n");
   return -4;
 error_hashdb_remove:
   fprintf(stderr, "error: cannot delete old hashdb '%s': %s\n", dbname, strerror(errno));
-  remove(dbtemp);
+  jc_remove(dbtemp);
   return -5;
 error_hashdb_rename:
   fprintf(stderr, "error: cannot rename temporary hashdb '%s' to '%s'; leaving it alone: %s\n", dbtemp, dbname, strerror(errno));
@@ -130,13 +131,14 @@ static int write_hashdb_entry(FILE *db, hashdb_t *cur, uint64_t *cnt, const int
 {
   struct timeval tm;
   int err = 0;
-  static char out[PATH_MAX + 128];
+  static char out[JC_PATHBUF_SIZE + 128];
 
+  LOUD(fprintf(stderr, "write_hashdb_entry(%p, %p, %p, %d)", db, cur, cnt, destroy);)
   /* Write header and traverse array on first call */
   if (unlikely(cur == NULL)) {
     if (hashdb_dirty == 1) {
       gettimeofday(&tm, NULL);
-      snprintf(out, PATH_MAX + 127, "jdupes hashdb:%d,%d,%08lx\n", HASHDB_VER, hash_algo, (unsigned long)tm.tv_sec);
+      snprintf(out, JC_PATHBUF_SIZE + 127, "jdupes hashdb:%d,%d,%08lx\n", HASHDB_VER, hash_algo, (unsigned long)tm.tv_sec);
       LOUD(fprintf(stderr, "write hashdb: %s", out);)
       errno = 0;
       fputs(out, db);
@@ -157,50 +159,31 @@ static int write_hashdb_entry(FILE *db, hashdb_t *cur, uint64_t *cnt, const int
 
   /* Write out this node if it wasn't invalidated */
   if (hashdb_dirty == 1 && cur->hashcount != 0) {
-    snprintf(out, PATH_MAX + 127, "%u,%016" PRIx64 ",%016" PRIx64 ",%016" PRIx64 ",%016" PRIx64 ",%016" PRIx64 ",%s\n",
+    snprintf(out, JC_PATHBUF_SIZE + 127, "%u,%016" PRIx64 ",%016" PRIx64 ",%016" PRIx64 ",%016" PRIx64 ",%016" PRIx64 ",%s\n",
       cur->hashcount, cur->partialhash, cur->fullhash, (uint64_t)cur->mtime, (uint64_t)cur->size, (uint64_t)cur->inode, cur->path);
     (*cnt)++;
     LOUD(fprintf(stderr, "write hashdb: %s", out);)
     errno = 0;
-    fputs(out, db);
+    if (db == NULL) printf("%s", out); else fputs(out, db);
     if (errno != 0) return 1;
   }
 
   /* Traverse the tree, propagating errors */
-  if (cur->left != NULL) err = write_hashdb_entry(db, cur->left, cnt, destroy);
+  if (err == 0 && cur->left != NULL) err = write_hashdb_entry(db, cur->left, cnt, destroy);
   if (err == 0 && cur->right != NULL) err = write_hashdb_entry(db, cur->right, cnt, destroy);
-  if (destroy == 1) { free(cur); }
+  if (destroy == 1) free(cur);
   return err;
 }
 
 
-#if 0
-void dump_hashdb(hashdb_t *cur)
+uint64_t dump_hashdb(void)
 {
-  struct timeval tm;
+  uint64_t cnt = 0;
 
-  if (cur == NULL) {
-    gettimeofday(&tm, NULL);
-    printf("jdupes hashdb:1,%d,%08lx\n", hash_algo, tm.tv_sec);
-    for (int i = 0; i < HT_SIZE; i++) {
-      if (hashdb[i] == NULL) continue;
-      dump_hashdb(hashdb[i]);
-    }
-    return;
-  }
-  /* db line format: hashcount,partial,full,mtime,path */
-#ifdef ON_WINDOWS
-  if (cur->hashcount != 0) printf("%u,%016llx,%016llx,%08llx,%08llx,%016llx,%s\n",
-      cur->hashcount, cur->partialhash, cur->fullhash, cur->mtime, cur->size, cur->inode, cur->path);
-#else
-  if (cur->hashcount != 0) printf("%u,%016lx,%016lx,%08lx,%08lx,%016lx,%s\n",
-      cur->hashcount, cur->partialhash, cur->fullhash, cur->mtime, cur->size, cur->inode, cur->path);
-#endif
-  if (cur->left != NULL) dump_hashdb(cur->left);
-  if (cur->right != NULL) dump_hashdb(cur->right);
-  return;
+  fprintf(stderr, "Dumping hash database\n");
+  write_hashdb_entry(NULL, NULL, &cnt, 0);
+  return cnt;
 }
-#endif
 
 
 static void pivot_hashdb_tree(hashdb_t **parent, hashdb_t *cur, enum pivot direction)
@@ -244,7 +227,7 @@ static void rebalance_hashdb_tree(hashdb_t **parent)
 }
 
 
-/* in_pathlen allows use of a precomputed path length to avoid extra strlen() calls */
+/* in_path allows use of a precomputed path length to avoid extra strlen() calls */
 hashdb_t *add_hashdb_entry(char *in_path, int pathlen, const file_t *check)
 {
   unsigned int bucket;
@@ -360,11 +343,11 @@ hashdb_t *add_hashdb_entry(char *in_path, int pathlen, const file_t *check)
 
 /* db header format: jdupes hashdb:dbversion,hashtype,update_mtime
  * db line format: hashcount,partial,full,mtime,size,inode,path */
-int64_t load_hash_database(char *dbname)
+int64_t load_hash_database(const char * const restrict dbname)
 {
   FILE *db;
-  char line[PATH_MAX + 128];
-  char buf[PATH_MAX + 128];
+  char line[JC_PATHBUF_SIZE + 128];
+  char buf[JC_PATHBUF_SIZE + 128];
   char *field, *temp;
   int db_ver;
   unsigned int fixed_len;
@@ -373,15 +356,15 @@ int64_t load_hash_database(char *dbname)
   time_t db_mtime;
   char date[32];
 #endif /* LOUD_DEBUG */
-  
+
   if (dbname == NULL) goto error_hashdb_null;
   LOUD(fprintf(stderr, "load_hash_database('%s')\n", dbname);)
   errno = 0;
-  db = fopen(dbname, "rb");
+  db = jc_fopen(dbname, JC_FILE_MODE_RDONLY_SEQ);
   if (db == NULL) goto warn_hashdb_open;
 
   /* Read header line */
-  if ((fgets(buf, PATH_MAX + 127, db) == NULL) || (ferror(db) != 0)) {
+  if ((fgets(buf, JC_PATHBUF_SIZE + 127, db) == NULL) || (ferror(db) != 0)) {
     if (errno == 0) goto warn_hashdb_open;  // empty file = make new DB
     goto error_hashdb_read;
   } else if (!ISFLAG(flags, F_HIDEPROGRESS)) fprintf(stderr, "Loading hash database...");
@@ -417,12 +400,12 @@ int64_t load_hash_database(char *dbname)
     jdupes_ino_t inode;
 
     errno = 0;
-    if ((fgets(line, PATH_MAX + 128, db) == NULL)) {
+    if ((fgets(line, JC_PATHBUF_SIZE + 128, db) == NULL)) {
       if (ferror(db) != 0) goto error_hashdb_read;
       break;
     }
     LOUD(fprintf(stderr, "read hashdb: %s", line);)
-    strncpy(buf, line, PATH_MAX + 128);
+    strncpy(buf, line, JC_PATHBUF_SIZE + 128);
     linenum++;
     linelen = (int64_t)strlen(buf);
     if (linelen < fixed_len + 1) goto error_hashdb_line;
@@ -445,9 +428,10 @@ int64_t load_hash_database(char *dbname)
     inode = strtoull(field, NULL, 16);
 
     path = buf + fixed_len;
+    path = remove_leading_dotslashes(path);
     path = strtok(path, "\n"); if (path == NULL) goto error_hashdb_line;
-    pathlen = linelen - fixed_len - 1;
-    if (pathlen > PATH_MAX) goto error_hashdb_line;
+    pathlen = linelen - fixed_len + 1;
+    if (pathlen > JC_PATHBUF_SIZE) goto error_hashdb_line;
     *(path + pathlen) = '\0';
 
     /* Allocate and populate a tree entry */
@@ -462,7 +446,7 @@ int64_t load_hash_database(char *dbname)
     entry->hashcount = hashcount;
   }
 
-  fclose(db);
+  jc_fclose(db);
   return linenum - 1;
 
 warn_hashdb_open:
@@ -471,50 +455,50 @@ warn_hashdb_open:
   return 0;
 error_hashdb_read:
   fprintf(stderr, "error reading hash database '%s': %s\n", dbname, strerror(errno));
-  fclose(db);
+  jc_fclose(db);
   return -1;
 error_hashdb_header:
   fprintf(stderr, "error in header of hash database '%s'\n", dbname);
-  fclose(db);
+  jc_fclose(db);
   return -2;
 error_hashdb_version:
   fprintf(stderr, "error: bad db version %u in hash database '%s'\n", db_ver, dbname);
-  fclose(db);
+  jc_fclose(db);
   return -3;
 error_hashdb_line:
   fprintf(stderr, "\nerror: bad line %" PRId64 " in hash database '%s':\n\n%s\n\n", linenum, dbname, line);
-  fclose(db);
+  jc_fclose(db);
   return -4;
 error_hashdb_add:
   fprintf(stderr, "error: internal failure allocating a hashdb entry\n");
-  fclose(db);
+  jc_fclose(db);
   return -5;
 error_hashdb_null:
   fprintf(stderr, "error: internal failure: NULL pointer for hashdb\n");
   return -6;
 warn_hashdb_algo:
   fprintf(stderr, "warning: hashdb uses a different hash algorithm than selected; not loading\n");
-  fclose(db);
+  jc_fclose(db);
   return -7;
 }
- 
+
 
 static int get_path_hash(char *path, int pathlen, uint64_t *path_hash)
 {
-  uint64_t aligned_path[(PATH_MAX + 8) / sizeof(uint64_t)];
+  uint64_t aligned_path[(JC_PATHBUF_SIZE + 8) / sizeof(uint64_t)];
   int retval;
 
   *path_hash = 0;
   if (pathlen < 1) pathlen = strlen(path);
   if ((uintptr_t)path & 0x0f) {
-    strncpy((char *)&aligned_path, path, PATH_MAX);
-    retval = jc_block_hash((uint64_t *)aligned_path, path_hash, pathlen);
-  } else retval = jc_block_hash((uint64_t *)path, path_hash, pathlen);
+    strncpy((char *)&aligned_path, path, JC_PATHBUF_SIZE);
+    retval = jc_block_hash(NORMAL, (uint64_t *)aligned_path, path_hash, strlen((char *)aligned_path));
+  } else retval = jc_block_hash(NORMAL, (uint64_t *)path, path_hash, strlen(path));
   return retval;
 }
 
 
- /* If file hash info is already present in hash database then preload those hashes */
+/* Scan database for a matching file entry; if found, load hashes into it */
 int read_hashdb_entry(file_t *file)
 {
   unsigned int bucket;
@@ -569,3 +553,56 @@ error_path_hash:
   fprintf(stderr, "error: internal error hashing a path\n");
   return -255;
 }
+
+
+int cleanup_hashdb(uint64_t *cnt, hashdb_t *cur)
+{
+  int err = 0;
+  static char **list;
+  static uint64_t listsize;
+  char *temp;
+  unsigned int i;
+
+  /* First call: traverse array, sort results, do actual verification */
+  if (unlikely(cur == NULL)) {
+    list = NULL;
+    listsize = 0;
+    *cnt = 0;
+
+    for (i = 0; i < HT_SIZE; i++) {
+      if (hashdb[i] == NULL) continue;
+      err = cleanup_hashdb(cnt, hashdb[i]);
+      if (err != 0) return err;
+    }
+
+    /* TODO: sort list first */
+
+    /* Check each item for existence; remove if it can't be accessed */
+    for (i = 0; i < *cnt; i++) {
+      char *path = *(list + i);
+      if (jc_access(path, JC_F_OK) == 0) continue;
+      /* TODO: invalidate entry - maybe need pointer? */
+    }
+
+    return 0;
+  }
+
+  /* If node is valid, add file to list to be checked, expanding array as needed */
+  if (cur->hashcount != 0) {
+    if (listsize == *cnt) {
+      listsize += 4096;
+      list = realloc(list, sizeof(char *) * listsize);
+      if (list == NULL) jc_oom("cleanup_hashdb realloc");
+    }
+    temp = (char *)malloc(strlen(cur->path) + 1);
+    if (temp == NULL) jc_oom("cleanup_hashdb path");
+    *(list + *cnt) = temp;
+    strcpy(temp, cur->path);
+    (void)*cnt++;
+  }
+
+  /* Traverse the tree, propagating errors */
+  if (err == 0 && cur->left != NULL) err = cleanup_hashdb(cnt, cur->left);
+  if (err == 0 && cur->right != NULL) err = cleanup_hashdb(cnt, cur->right);
+  return err;
+}
