@@ -1,5 +1,5 @@
 /* jdupes duplicate file finder utility
- * Copyright (C) 2015-2023 by Jody Bruchon <jody@jodybruchon.com>
+ * Copyright (C) 2015-2025 by Jody Bruchon <jody@jodybruchon.com>
  * Licensed under The MIT License; see LICENSE.txt for details */
 
 #include <stdio.h>
@@ -72,13 +72,8 @@
 /* Detect Windows and modify as needed */
 #if defined _WIN32 || defined __MINGW32__
  #ifdef UNICODE
-  const wchar_t *FILE_MODE_RO = L"rbS";
-  wpath_t wstr;
- #else
-  const char *FILE_MODE_RO = "rbS";
  #endif /* UNICODE */
 #else /* Not Windows */
- const char *FILE_MODE_RO = "rb";
  #ifdef UNICODE
   #error Do not define UNICODE on non-Windows platforms.
   #undef UNICODE
@@ -90,19 +85,19 @@ uint_fast64_t flags = 0, a_flags = 0, p_flags = 0;
 
 static const char *program_name;
 
-/* Stat and SIGUSR */
-#ifdef ON_WINDOWS
- struct jc_winstat s;
-#else
- struct stat s;
-#endif
-
 #ifndef PARTIAL_HASH_SIZE
  #define PARTIAL_HASH_SIZE 4096
 #endif
 
 #ifndef NO_CHUNKSIZE
  size_t auto_chunk_size = CHUNK_SIZE;
+ #define MIN_CHUNK_SIZE 4096
+ #define MAX_CHUNK_SIZE (1048576 * 1024)
+ #define MAX_CHUNK_OPT_SIZE 18
+ /* Chunk size is now exponential to make things easier */
+ static unsigned int chunk_opt_size[] = {
+    1 << 12, 1 << 13, 1 << 14, 1 << 15, 1 << 16, 1 << 17, 1 << 18, 1 << 19, 1 << 20, 1 << 21, 1 << 22, 1 << 23, 1 << 24, 1 << 25, 1 << 26, 1 << 27, 1 << 28, 1 << 29, 1 << 30
+ };
 #else
  /* If automatic chunk sizing is disabled, just use a fixed value */
  #define auto_chunk_size CHUNK_SIZE
@@ -140,11 +135,13 @@ unsigned int user_item_count = 1;
 int sort_direction = 1;
 
 /* For path name mangling */
-char tempname[PATHBUF_SIZE * 2];
+char tempname[JC_PATHBUF_SIZE * 2];
 
 /* Strings used in multiple places */
 const char *s_interrupt = "\nStopping file scan due to user abort\n";
 const char *s_no_dupes = "No duplicates found.\n";
+const char *s_help = "help";
+const char *s_chunksize_help = "         Run 'jdupes -C help' or info on the new size specifications.\n\n";
 
 /* Exit status; use exit() codes for setting this */
 int exit_status = EXIT_SUCCESS;
@@ -172,9 +169,7 @@ int main(int argc, char **argv)
 #endif
 #ifndef NO_CHUNKSIZE
   static long manual_chunk_size = 0;
- #ifdef __linux__
-  static struct jc_proc_cacheinfo pci;
- #endif /* __linux__ */
+  static struct jc_proc_cacheinfo *pci;
 #endif /* NO_CHUNKSIZE */
 #ifdef ENABLE_DEDUPE
  #ifdef __linux__
@@ -231,6 +226,7 @@ int main(int argc, char **argv)
     { "version", 0, 0, 'v' },
     { "ext-filter", 1, 0, 'X' },
     { "hash-db", 1, 0, 'y' },
+    { "hash-db-populate", 1, 0, 'Y' },
     { "soft-abort", 0, 0, 'Z' },
     { "zero-match", 0, 0, 'z' },
     { NULL, 0, 0, 0 }
@@ -240,7 +236,7 @@ int main(int argc, char **argv)
  #define GETOPT getopt
 #endif
 
-#define GETOPT_STRING "@019ABC:DdEefHhIijKLlMmNnOo:P:pQqRrSsTtUuVvX:y:Zz"
+#define GETOPT_STRING "@019ABC:DdEefHhIijKLlMmNnOo:P:pQqRrSsTtUuVvX:Yy:Zz"
 
   /* Verify libjodycode compatibility before going further */
   if (libjodycode_version_check(1, 0) != 0) {
@@ -260,28 +256,26 @@ int main(int argc, char **argv)
   int wa_err;
   argv = (char **)malloc(sizeof(char *) * (size_t)argc);
   if (!argv) jc_oom("main() unicode argv");
-  wa_err = jc_widearg_to_argv(argc, wargv, argv);
+  wa_err = jc_widearg_to_argv(argc, wargv, &argv);
   if (wa_err != 0) {
     jc_print_error(wa_err);
     exit(EXIT_FAILURE);
   }
-  /* fix up __argv so getopt etc. don't crash */
-  __argv = argv;
-  jc_set_output_modes(0x0c);
+  jc_set_output_modes(JC_MODE_UTF16_TTY, JC_MODE_UTF16_TTY);
 #endif /* UNICODE */
 
 #ifndef NO_CHUNKSIZE
-#ifdef __linux__
   /* Auto-tune chunk size to be half of L1 data cache if possible */
-  jc_get_proc_cacheinfo(&pci);
-  if (pci.l1 != 0) auto_chunk_size = (pci.l1 / 2);
-  else if (pci.l1d != 0) auto_chunk_size = (pci.l1d / 2);
-  /* Must be at least 4096 (4 KiB) and cannot exceed CHUNK_SIZE */
-  if (auto_chunk_size < MIN_CHUNK_SIZE || auto_chunk_size > MAX_CHUNK_SIZE) auto_chunk_size = CHUNK_SIZE;
-  /* Force to a multiple of 4096 if it isn't already */
-  if ((auto_chunk_size & 0x00000fffUL) != 0)
-    auto_chunk_size = (auto_chunk_size + 0x00000fffUL) & 0x000ff000;
-#endif /* __linux__ */
+  pci = jc_get_proc_cacheinfo(0);
+  if (pci != NULL) {
+    if (pci->l1 != 0) auto_chunk_size = (pci->l1 / 2);
+    else if (pci->l1d != 0) auto_chunk_size = (pci->l1d / 2);
+    /* Must be at least 4096 (4 KiB) and cannot exceed MAX_CHUNK_SIZE */
+    if (auto_chunk_size < MIN_CHUNK_SIZE || auto_chunk_size > MAX_CHUNK_SIZE) auto_chunk_size = CHUNK_SIZE;
+    /* Force to a multiple of 4096 if it isn't already */
+    if ((auto_chunk_size & 0x00000fffUL) != 0)
+      auto_chunk_size = (auto_chunk_size + 0x00000fffUL) & 0x000ff000;
+  }
 #endif /* NO_CHUNKSIZE */
 
   /* Is stderr a terminal? If not, we won't write progress to it */
@@ -341,13 +335,20 @@ int main(int argc, char **argv)
 #endif /* ENABLE_DEDUPE */
 #ifndef NO_CHUNKSIZE
     case 'C':
-      manual_chunk_size = (strtol(optarg, NULL, 10) & 0x0ffffffcL) << 10;  /* Align to 4K sizes */
-      if (manual_chunk_size < MIN_CHUNK_SIZE || manual_chunk_size > MAX_CHUNK_SIZE) {
-        fprintf(stderr, "warning: invalid manual chunk size (must be %d - %d KiB); using defaults\n", MIN_CHUNK_SIZE / 1024, MAX_CHUNK_SIZE / 1024);
-        LOUD(fprintf(stderr, "Manual chunk size (failed) was apparently '%s' => %ld KiB\n", optarg, manual_chunk_size / 1024));
-        manual_chunk_size = 0;
-      } else auto_chunk_size = (size_t)manual_chunk_size;
-      LOUD(fprintf(stderr, "Manual chunk size is %ld\n", manual_chunk_size));
+      errno = 0;
+      if (jc_strcaseeq(optarg, s_help) == 0) { help_text_chunksize(); exit(EXIT_SUCCESS); }
+      manual_chunk_size = strtol(optarg, NULL, 10);
+      if (errno != 0 || manual_chunk_size < 0 || manual_chunk_size > MAX_CHUNK_OPT_SIZE) {
+        fprintf(stderr, "error: -C/--chunk-size '%s' invalid or out of range\n", optarg);
+        fprintf(stderr, "%s", s_chunksize_help);
+	exit(EXIT_FAILURE);
+      }
+      fprintf(stderr, "\nWARNING: The -C/--chunk-size parameter has changed!\n");
+      fprintf(stderr, "%s", s_chunksize_help);
+      auto_chunk_size = chunk_opt_size[manual_chunk_size];
+      fprintf(stderr, "note: the user-selected I/O chunk size is ");
+      if (auto_chunk_size <= 524288) fprintf(stderr, "%lld KiB\n", auto_chunk_size / 1024);
+      if (auto_chunk_size > 524288) fprintf(stderr, "%lld MiB\n", auto_chunk_size / 1048576);
       break;
 #endif /* NO_CHUNKSIZE */
 #ifndef NO_DELETE
@@ -516,6 +517,8 @@ int main(int argc, char **argv)
       break;
 #ifndef NO_EXTFILTER
     case 'X':
+      /* Invoke help text if requested */
+      if (jc_strcaseeq(optarg, s_help) == 0) { help_text_extfilter(); exit(EXIT_SUCCESS); }
       add_extfilter(optarg);
       break;
 #endif /* NO_EXTFILTER */
@@ -535,6 +538,9 @@ int main(int argc, char **argv)
       if (strcmp(optarg, ".") == 0) strcpy(hashdb_name, "jdupes_hashdb.txt");
       else strcpy(hashdb_name, optarg);
       break;
+    case 'Y':
+      SETFLAG(flags, F_HASHDB_POPULATE);
+      break;
 #endif /* NO_HASHDB */
     case 'z':
       SETFLAG(flags, F_INCLUDEEMPTY);
@@ -558,6 +564,13 @@ int main(int argc, char **argv)
     }
   }
 
+#ifndef NO_HASHDB
+  if (ISFLAG(flags, F_HASHDB_POPULATE)) {
+    if (!ISFLAG(flags, F_HASHDB)) goto error_hashdb_populate;
+    if (!ISFLAG(flags, F_HIDEPROGRESS)) fprintf(stderr, "Pre-populating hash database only. Duplicate detection is disabled.\n");
+  }
+#endif
+
   if (optind >= argc) {
     fprintf(stderr, "no files or directories specified (use -h option for help)\n");
     exit(EXIT_FAILURE);
@@ -661,7 +674,6 @@ skip_partialonly_noise:
     /* F_RECURSE is not set for directories before --recurse: */
     for (int x = optind; x < firstrecurse; x++) {
       if (unlikely(interrupt)) goto interrupt_exit;
-      jc_slash_convert(argv[x]);
       loaddir(argv[x], &files, 0);
       user_item_count++;
     }
@@ -671,14 +683,12 @@ skip_partialonly_noise:
 
     for (int x = firstrecurse; x < argc; x++) {
       if (unlikely(interrupt)) goto interrupt_exit;
-      jc_slash_convert(argv[x]);
       loaddir(argv[x], &files, 1);
       user_item_count++;
     }
   } else {
     for (int x = optind; x < argc; x++) {
       if (unlikely(interrupt)) goto interrupt_exit;
-      jc_slash_convert(argv[x]);
       loaddir(argv[x], &files, ISFLAG(flags, F_RECURSE));
       user_item_count++;
     }
@@ -715,11 +725,8 @@ skip_partialonly_noise:
 
   while (curfile) {
     static file_t **match = NULL;
-    static FILE *file1;
-    static FILE *file2;
 
-    if (interrupt) {
-      fprintf(stderr, "%s", s_interrupt);
+    if (unlikely(interrupt != 0)) {
       if (!ISFLAG(flags, F_SOFTABORT)) exit(EXIT_FAILURE);
       interrupt = 0;  /* reset interrupt for re-use */
       goto skip_file_scan;
@@ -727,6 +734,33 @@ skip_partialonly_noise:
 
     LOUD(fprintf(stderr, "\nMAIN: current file: %s\n", curfile->d_name));
 
+    /* If -Y passed, populate the hash database with this file and do nothing else */
+#ifndef NO_HASHDB
+    if (unlikely(ISFLAG(flags, F_HASHDB_POPULATE))) {
+      const uint64_t * restrict pop_filehash;
+      int pop_dirty;
+      pop_dirty = 0;
+      if (!ISFLAG(curfile->flags, FF_HASH_PARTIAL)) {
+        pop_filehash = get_filehash(curfile, PARTIAL_HASH_SIZE, hash_algo);
+        if (pop_filehash == NULL) goto skip_full_check;
+        curfile->filehash_partial = *pop_filehash;
+	pop_dirty = 1;
+      }
+      if (!ISFLAG(curfile->flags, FF_HASH_FULL)) {
+        pop_filehash = get_filehash(curfile, 0, hash_algo);
+        if (pop_filehash == NULL) goto skip_full_check;
+        curfile->filehash = *pop_filehash;
+	pop_dirty = 1;
+      }
+      if (pop_dirty == 1) {
+        SETFLAG(curfile->flags, FF_HASH_PARTIAL | FF_HASH_FULL | FF_HASHDB_DIRTY);
+        add_file_to_hashdb(curfile);
+      }
+      goto skip_full_check;
+    }
+#endif
+
+    /* Go through the tree seeing if anything matches the current file */
     if (!checktree) registerfile(&checktree, NONE, curfile);
     else match = checkmatch(checktree, curfile);
 
@@ -754,32 +788,7 @@ skip_partialonly_noise:
         goto skip_full_check;
       }
 
-#ifdef UNICODE
-      if (!M2W(curfile->d_name, wstr)) file1 = NULL;
-      else file1 = _wfopen(wstr, FILE_MODE_RO);
-#else
-      file1 = fopen(curfile->d_name, FILE_MODE_RO);
-#endif
-      if (!file1) {
-        LOUD(fprintf(stderr, "MAIN: warning: file1 fopen() failed ('%s')\n", curfile->d_name));
-        curfile = curfile->next;
-        continue;
-      }
-
-#ifdef UNICODE
-      if (!M2W((*match)->d_name, wstr)) file2 = NULL;
-      else file2 = _wfopen(wstr, FILE_MODE_RO);
-#else
-      file2 = fopen((*match)->d_name, FILE_MODE_RO);
-#endif
-      if (!file2) {
-        fclose(file1);
-        LOUD(fprintf(stderr, "MAIN: warning: file2 fopen() failed ('%s')\n", (*match)->d_name));
-        curfile = curfile->next;
-        continue;
-      }
-
-      if (confirmmatch(file1, file2, curfile->size)) {
+      if (confirmmatch(curfile->d_name, (*match)->d_name, curfile->size) == 0) {
         LOUD(fprintf(stderr, "MAIN: registering matched file pair\n"));
 #ifndef NO_MTIME
         registerpair(match, curfile, (ordertype == ORDER_TIME) ? sort_pairs_by_mtime : sort_pairs_by_filename);
@@ -787,10 +796,10 @@ skip_partialonly_noise:
         registerpair(match, curfile, sort_pairs_by_filename);
 #endif
         dupecount++;
-      } DBG(else hash_fail++;)
-
-      fclose(file1);
-      fclose(file2);
+      } else {
+	goto skip_full_check;
+        DBG(hash_fail++;)
+      }
     }
 
 skip_full_check:
@@ -811,6 +820,10 @@ skip_file_scan:
   signal(SIGINT, SIG_DFL);
   if (!ISFLAG(flags, F_HIDEPROGRESS)) jc_stop_alarm();
 
+#ifndef NO_HASHDB
+  if (ISFLAG(flags, F_HASHDB_POPULATE)) goto skip_all_actions;
+#endif
+
   if (files == NULL) {
     printf("%s", s_no_dupes);
     exit(exit_status);
@@ -841,7 +854,9 @@ skip_file_scan:
     summarizematches(files);
   }
 
+
 #ifndef NO_HASHDB
+skip_all_actions:
   if (ISFLAG(flags, F_HASHDB)) {
     hdbout = save_hash_database(hashdb_name, 1);
     if (!ISFLAG(flags, F_HIDEPROGRESS)) {
@@ -866,7 +881,9 @@ skip_all_scan_code:
     if (manual_chunk_size > 0) fprintf(stderr, "I/O chunk size: %ld KiB (manually set)\n", manual_chunk_size >> 10);
     else {
   #ifdef __linux__
-      fprintf(stderr, "I/O chunk size: %" PRIuMAX " KiB (%s)\n", (uintmax_t)(auto_chunk_size >> 10), (pci.l1 + pci.l1d) != 0 ? "dynamically sized" : "default size");
+      if (pci != NULL) {
+        fprintf(stderr, "I/O chunk size: %" PRIuMAX " KiB (%s)\n", (uintmax_t)(auto_chunk_size >> 10), (pci->l1 + pci->l1d) != 0 ? "dynamically sized" : "default size");
+      }
   #else
       fprintf(stderr, "I/O chunk size: %" PRIuMAX " KiB (default size)\n", (uintmax_t)(auto_chunk_size >> 10));
   #endif /* __linux__ */
@@ -890,6 +907,9 @@ error_optarg:
 error_load_hashdb:
   free(hashdb_name);
   exit(EXIT_FAILURE);
+error_hashdb_populate:
+  fprintf(stderr, "error: -Y/--hash-db-populate requires the -y/--hash-db option\n");
+  exit(EXIT_FAILURE);
 #endif
 interrupt_exit:
   fprintf(stderr, "%s", s_interrupt);
